/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "MeshFInfo.hpp"

namespace parncl{

const std::string MOAB_FLOAD_OPTIONS("PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;PARALLEL_GHOSTS=2.0.1;PARTITION_METHOD=");
const std::string MOAB_VLOAD_OPTIONS(";VARIABLE=");


MeshFInfo::MeshFInfo(std::vector<std::string> &fnames, moab::Interface *mb_gcore, moab::ParallelComm *mb_pcomm, const std::string &part_method):mb_gcore_(mb_gcore), fnames_(fnames), next_dim_id_(0)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

    mb_retval = mb_gcore->create_meshset(moab::MESHSET_SET, mb_mset_);
    if(mb_retval != moab::MB_SUCCESS){
        std::string funcname("UNKNOWN");
        NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to create MOAB meshset");
        throw pargal::pargal_except(mb_gcore->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
    }

    std::string mb_fload_options = MOAB_FLOAD_OPTIONS + part_method + MOAB_VLOAD_OPTIONS;
    
    mb_retval = mb_gcore->load_file(fnames[0].c_str(), &mb_mset_, mb_fload_options.c_str());
    if(mb_retval != moab::MB_SUCCESS){
        mb_gcore->clear_meshset(&mb_mset_, 1);
        std::string funcname("UNKNOWN");
        NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to load file");
        throw pargal::pargal_except(mb_gcore->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
    }

	pcal_finfo_ = new pargal::fileinfo(fnames, *mb_gcore, *mb_pcomm, mb_mset_, part_method);
}

MeshFInfo::~MeshFInfo()
{
    mb_gcore_->clear_meshset(&mb_mset_, 1);
	if(pcal_finfo_){
		delete pcal_finfo_;
	}
}

moab::EntityHandle MeshFInfo::get_mb_mset(void ) const
{
	return mb_mset_;
}

pargal::fileinfo &MeshFInfo::get_pcal_finfo(void ) const
{
	return (*pcal_finfo_);
}

size_t MeshFInfo::get_num_nc_files(void ) const
{
	return fnames_.size();
} 

const std::vector<std::string> &MeshFInfo::get_nc_fnames(void ) const
{
	return fnames_;
}

size_t MeshFInfo::insert_dim_name(const std::string &dim_name)
{
	size_t dim_id = next_dim_id_++;		
	dims_map_.insert(std::make_pair(std::string(dim_name), dim_id));
	return dim_id;
}

size_t MeshFInfo::get_dim_id(const std::string &dim_name)
{
	size_t dim_id = FINFO_INVALID_DIM_ID;
	std::map<std::string, size_t>::const_iterator iter = dims_map_.find(dim_name);
	if(iter != dims_map_.end()){
		dim_id = (*iter).second;	
	}
	else{
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get dim_id for %s", dim_name.c_str());
	}
	return dim_id;
}

} // namespace parncl
