/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "MBMesh.hpp"
#include "ParCALMeshVal.hpp"
#include "PNclMBUtils.hpp"
#include <iostream>

namespace parncl{

// ParCALMeshVal Sptr func defns
ParCALMeshValSPtr::ParCALMeshValSPtr(ParCALMeshVal *valp):valp_(valp)
{
	valp_->inc_ref_cnt();
}

ParCALMeshVal *ParCALMeshValSPtr::operator->()
{
	// Access mesh val via sptr
	return valp_;
}

ParCALMeshVal &ParCALMeshValSPtr::operator*()
{
	// Copy mesh val by value
	if(valp_->is_file_val() || (valp_->get_ref_cnt() > 1)){
		ParCALMeshVal *new_val = new ParCALMeshVal(*valp_);
		return *new_val;
	}
	return *valp_;
}

ParCALMeshValSPtr::ParCALMeshValSPtr(const ParCALMeshValSPtr &sptr):valp_(sptr.valp_)
{
	// Copy by ref
	valp_->inc_ref_cnt();
}

ParCALMeshValSPtr& ParCALMeshValSPtr::operator=(const ParCALMeshValSPtr &rhs)
{
	if(this != &rhs){
		valp_->dec_ref_cnt();
		if(valp_->get_ref_cnt() == 0){
			delete valp_;
		}
		valp_ = rhs.valp_;
		valp_->inc_ref_cnt();
	}
	return *this;
}

ParCALMeshValSPtr::~ParCALMeshValSPtr()
{
	valp_->dec_ref_cnt();
	if(valp_->get_ref_cnt() == 0){
		delete valp_;
	}
}

// ParCALMeshVal func defns ....

std::string ParCALMeshVal::convert_to_parcal_loc(const ParCALMeshVal::MeshValLoc val_loc)
{
	switch(val_loc){
		case MESHVAL_ON_QUAD_CENTER:
				return std::string("QUAD");
		default:
				return std::string("UNIMPL");	
	}
}

ParCALMeshVal::ParCALMeshVal(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar):prev_cached_lval_chunk_(parncl::PARCALVAL_NCHUNK_NONE), prev_cached_lval_buf_(NULL), ncached_lval_chunks_(0), mesh_(NULL), mb_core_(NULL), val_loc_(ParCALMeshVal::MESHVAL_LOC_UNKNOWN), pcvar_(pcvar), data_type_(data_type), num_pcvar_chunks_(0), pcvar_chunk_sz_(0), num_pcvar_levs_(0), is_fvar_(is_fvar), ref_cnt_(0)
{
	cached_gval_ = std::make_pair(false, (std::vector<double> *)NULL);
    const std::vector<pargal::pcdim>& pcvar_dims = pcvar->get_dims();
	for(std::vector<pargal::pcdim>::const_iterator iter = pcvar_dims.begin(); iter != pcvar_dims.end(); ++iter){
		if((*iter).name() == std::string("lev")){
			num_pcvar_levs_ = (*iter).size();
			break;
		}
		if((*iter).name() == std::string("ilev")){
			num_pcvar_levs_ = (*iter).size();
			break;
		}
	}
}

ParCALMeshValSPtr ParCALMeshVal::create_mesh_val(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar)
{
	ParCALMeshVal *mesh_val = new ParCALMeshVal(pcvar, data_type, is_fvar);
	mesh_val->num_pcvar_chunks_ = pcvar->get_tsteps().size();
	return (ParCALMeshValSPtr(mesh_val));
}

ParCALMeshValSPtr ParCALMeshVal::create_mesh_val(MBMesh *mesh, moab::Interface *mb_core, size_t mb_chunk_sz, const std::vector<int> &val_tsteps, const std::vector<pargal::pcdim> &val_dims, size_t nlevs, nc_type data_type, const std::string &pcal_val_loc)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

	assert(mesh != NULL);
	MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle mb_set = pcal_finfo.get_file_set();
	moab::DataType dt = PNclNCMapToMB(data_type);
	std::string pcvar_name = ParCALGetUniqStr(mesh);

	for(std::vector<int>::const_iterator iter = val_tsteps.begin(); iter != val_tsteps.end(); ++iter){
		size_t tstep = *iter;
		moab::EntityType ent_type = ParCALLocToMB(pcal_val_loc);
		mb_retval = PNclMBCreateTag(mesh, pcvar_name.c_str(), tstep, nlevs, NULL, dt, ent_type);
		if(mb_retval != moab::MB_SUCCESS){
			std::string funcname("UNKNOWN");
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating MOAB tag while creating var");
			throw pargal::pargal_except(mb_core->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
		}
	}

	pargal::pcvar *pcvar = new pargal::pcvar(pcvar_name, dt,
									val_dims, pcal_finfo, val_tsteps, true, pcal_val_loc);

	ParCALMeshVal *mesh_val = new ParCALMeshVal(pcvar, data_type, false);
	mesh_val->num_pcvar_chunks_ = pcvar->get_tsteps().size();
	// FIXME: Find val loc from dim names
	mesh_val->set_mesh_info(mesh, mb_core, mb_chunk_sz, ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
	return (ParCALMeshValSPtr(mesh_val));
}

ParCALMeshVal::ParCALMeshVal(ParCALMeshVal &val):prev_cached_lval_chunk_(parncl::PARCALVAL_NCHUNK_NONE), prev_cached_lval_buf_(NULL), ncached_lval_chunks_(0), mesh_(val.mesh_), mb_core_(val.mb_core_), val_loc_(ParCALMeshVal::MESHVAL_LOC_UNKNOWN), pcvar_(NULL), data_type_(val.data_type_), num_pcvar_chunks_(val.num_pcvar_chunks_), pcvar_chunk_sz_(val.pcvar_chunk_sz_), is_fvar_(false), ref_cnt_(0)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	parncl::MeshFInfo &mesh_finfo = mesh_->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	const std::vector<int>& pcvar_tsteps = val.pcvar_->get_tsteps();
	if(val.is_fvar_){
		// Record the root variable in the file
		root_fvar_name_ = val.root_fvar_name_;
	}
	std::string pcvar_name = ParCALGetUniqStr(mesh_);
	moab::DataType dt = PNclNCMapToMB(data_type_);
	moab::EntityType ent_type = ParCALLocToMB(val.pcvar_->get_loc());
	size_t nlevs = val.num_pcvar_levs_;

    for(std::vector<int>::const_iterator iter = pcvar_tsteps.begin(); iter != pcvar_tsteps.end(); ++iter){
        size_t to_tstep = *iter, from_tstep = *iter;

        val.get_cached_lval(*iter);
        if(val.is_fvar_){
            if(mesh_finfo.get_num_nc_files() > 1){
                /* If time steps are spread across files, 
                 * MOAB assumes that only 1 timestep exists in a file and 
                 * that loads any time step as timestep 0
                 */
                from_tstep = 0;
            }
        }
        mb_retval = PNclMBDupTag(mesh_, pcvar_name.c_str(), to_tstep, val.pcvar_->name().c_str(), from_tstep, nlevs, dt, ent_type);
        if(mb_retval != moab::MB_SUCCESS){
			// FIXME: Is there a portable way to get this info
			std::string funcname("UNKNOWN");
            NhlPError(NhlFATAL, NhlEUNKNOWN, "Error duplicating MOAB tag");
			throw pargal::pargal_except(mb_core_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
        }
		val.free_cached_lval(*iter);
    }
	// Create a new pcvar for the val
    const std::vector<pargal::pcdim>& pcvar_dims = val.pcvar_->get_dims();
    pcvar_ = new pargal::pcvar(pcvar_name, dt, pcvar_dims, pcal_finfo, pcvar_tsteps, true, val.pcvar_->get_loc());
	// FIXME: We don't catch memalloc failure
	cached_gval_ = std::make_pair(false, (std::vector<double> *)NULL);
}

ParCALMeshVal::~ParCALMeshVal()
{
	// FIXME: Delete pcvar here, free cached global val
}

void ParCALMeshVal::set_mesh_info(MBMesh *mesh, moab::Interface *mb_core, size_t chunk_sz, ParCALMeshVal::MeshValLoc val_loc)
{
	mesh_ = mesh;
	mb_core_ = mb_core;
	pcvar_chunk_sz_ = chunk_sz * num_pcvar_levs_;
	val_loc_ = val_loc;
}

ParCALMeshVal::MeshValLoc ParCALMeshVal::get_mesh_val_loc(void ) const
{
	return val_loc_;
}

pargal::fileinfo &ParCALMeshVal::get_pcal_finfo(void ) const
{
	MeshFInfo &mesh_finfo = mesh_->get_mesh_finfo();
	return mesh_finfo.get_pcal_finfo();
}

MBMesh *ParCALMeshVal::get_mesh(void ) const
{
	return mesh_;
}

size_t ParCALMeshVal::get_num_val_chunks(void ) const
{
	return num_pcvar_chunks_;
}

size_t ParCALMeshVal::get_num_val_levs(void ) const
{
	return num_pcvar_levs_;
}

size_t ParCALMeshVal::get_val_chunk_sz(void ) const
{
//	return pcvar_chunk_sz_;
	return pcvar_->tstep_size();
}

// FIXME: Don't assume double
std::vector<double> *ParCALMeshVal::get_gval(void )
{
	if(cached_gval_.first){
		return cached_gval_.second;
	}
	// FIXME: Don't assume double
	std::vector<double> *gdata = new std::vector<double>();
	gdata->clear();
	pargal::gather(*pcvar_, pcvar_->get_tsteps(), *gdata);

	cached_gval_.second = gdata;
	cached_gval_.first = true;

	return cached_gval_.second;
}

void *ParCALMeshVal::get_cached_lval(size_t chunk)
{
	void *buf = NULL;
	// FIXME: Make MB design more flexible to get rid of this ugly assert
	if(ncached_lval_chunks_ == 0){
		buf = (void *)pcvar_->get_storage(chunk);
		prev_cached_lval_chunk_ = chunk;
		prev_cached_lval_buf_ = buf;
	}
	else{
		assert(chunk == prev_cached_lval_chunk_);
		buf = prev_cached_lval_buf_;
	}
	ncached_lval_chunks_++;
 	return buf;
}

void ParCALMeshVal::free_cached_lval(size_t chunk)
{
	ncached_lval_chunks_--;
	assert(chunk == prev_cached_lval_chunk_);
	if(is_fvar_ && (ncached_lval_chunks_ == 0)){
		pcvar_->delete_storage(prev_cached_lval_chunk_);	
		prev_cached_lval_chunk_ = parncl::PARCALVAL_NCHUNK_NONE;
		prev_cached_lval_buf_ = NULL;
	}
}

pargal::pcvar *ParCALMeshVal::get_pcvar(void )
{
	return pcvar_;
}

nc_type ParCALMeshVal::get_data_type(void ) const
{
	return data_type_;
}

bool ParCALMeshVal::is_file_val(void ) const
{
	return is_fvar_;
}

int ParCALMeshVal::get_ref_cnt(void ) const
{
	return ref_cnt_;
}

void ParCALMeshVal::inc_ref_cnt(void )
{
	ref_cnt_++;
}

void ParCALMeshVal::dec_ref_cnt(void)
{
	ref_cnt_--;
}

} // namespace parncl
