/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "ParNCLNetCdf.h"

typedef struct _PNNCOption{
	NclQuark name;
	NclBasicDataTypes data_type;
	int n_values;
	void *values;
} PNNCOption;

enum PNNCOptionsId{
	PNNC_SUPPRESS_CLOSE_OPT=0,
	/* PNNC_NUM_OPTIONS should be the last member of this enum */
	PNNC_NUM_OPTIONS
};

typedef struct _PNNCFileRecord {
NclQuark	file_path_q;
int		wr_status;
int		n_vars;
PNNCVarInqRecList *vars;
int		n_dims;
PNNCDimInqRecList *dims;
int		n_file_atts;
PNNCAttInqRecList *file_atts;
void *pcal_hnd;
int             n_options;
PNNCOption   *options;
} PNNCFileRecord;

static NclBasicDataTypes ParNCLNCMapToNcl(void *type);
/*FIXME: Get rid of this extern decl */
extern void *ParCALCacheVar(void **pcal_hnds, int nprs, NclQuark var_name, NclFVarSelRec *sel, int nsel);

static void *ParNCLNCInitializeFileRec(NclFileFormat *fmt)
{
	PNNCFileRecord *prec = NULL;

	prec = (PNNCFileRecord *)NclCalloc(1, sizeof(PNNCFileRecord));
	if(!prec){
		NhlPError(NhlFATAL, ENOMEM, NULL);
		return NULL;
	}

	/* Add default file options */
	prec->n_options = PNNC_NUM_OPTIONS;
	prec->options = NclCalloc(PNNC_NUM_OPTIONS, sizeof(PNNCOption));

	prec->options[PNNC_SUPPRESS_CLOSE_OPT].name = NrmStringToQuark("suppressclose");
	prec->options[PNNC_SUPPRESS_CLOSE_OPT].data_type = NCL_int;
	prec->options[PNNC_SUPPRESS_CLOSE_OPT].n_values = 1;
	prec->options[PNNC_SUPPRESS_CLOSE_OPT].values = NULL;

	*fmt = _ParNCLNETCDF;
	return ((void *) prec);
}

static void ParNCLNCFreeFileRec(void *pr)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;

	if(prec == NULL)
		return;

	if(prec->pcal_hnd != NULL){
		ParCALClose(prec->pcal_hnd, prec->file_path_q);
	}
	NclFree(prec);
}


static void *ParNCLNCCreateFile(void *pr, NclQuark path)
{
	/* FIXME: Add support for createfile() */
	NhlPError(NhlFATAL, NhlEUNKNOWN, "CreateFile() is not supported");
	return NULL;
}

static NhlErrorTypes ParNCLNCCreateFiles(void **prs, int nprs, NclMultiDValData paths)
{
	/* FIXME: Add support for createfile() */
	NhlPError(NhlFATAL, NhlEUNKNOWN, "CreateFiles() is not supported");
	return NhlFATAL;
}

static NhlErrorTypes ParNCLNCGetFileMData(PNNCFileRecord *prec)
{

	/*
	 * FIXME: It looks like we need to get DimInfo() before VarInfo()
	 * - Test to see if this is the case and fix it
	 */
	prec->n_dims = ParCALGetDimInfo(prec->pcal_hnd, &(prec->dims));
	if(prec->n_dims <= 0){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get dimension info from ParCAL");
		return NhlFATAL;
	}

	prec->n_vars = ParCALGetVarInfo(prec->pcal_hnd, &(prec->vars));
	if(prec->n_vars <= 0){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get variable info from ParCAL");
		return NhlFATAL;
	}

	prec->n_file_atts = ParCALGetFileAttInfo(prec->pcal_hnd, &(prec->file_atts));
	if(prec->n_file_atts < 0){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get attribute info from ParCAL");
		return NhlFATAL;
	}
	
	return NhlNOERROR;
}

static void *ParNCLNCOpenFile(void *pr, NclQuark path, int wr_status)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	NhlErrorTypes ret;
	if(prec == NULL)
		return NULL;
	prec->pcal_hnd = ParCALOpen(path, wr_status);
	if(prec->pcal_hnd == NULL){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "OpenFile() failed");
		return NULL;
	}
	
	ret = ParNCLNCGetFileMData(prec);
	if(ret < NhlNOERROR){
		NhlPError(NhlWARNING, NhlEUNKNOWN, "Reading meta data from file failed");
	}
	
	return prec;
}

static NhlErrorTypes ParNCLNCOpenFiles(void **prs, int nprs, NclMultiDValData paths, int wr_status)
{
	int i;
	void **pcal_hnds = NULL;
	NhlErrorTypes ret;

	/* FIXME: Avoid these unnecessary mallocs, maybe ParCALMOpen() should
	 * recognize parallel file records ?
	 * ParCALMOpen() recognizing PNNC file recs can be an issue if we need
	 * to add support for other file types in future
	 */
	pcal_hnds = NclCalloc(nprs, sizeof(void *));
	if(pcal_hnds == NULL){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "OpenFiles() failed");
		return NhlFATAL;
	}

	ret = ParCALMOpen(pcal_hnds, nprs, paths, wr_status);
	if(ret < NhlWARNING){
		NclFree(pcal_hnds);
		NhlPError(NhlFATAL, NhlEUNKNOWN, "OpenFiles() failed");
		return NhlFATAL;
	}

	for(i=0; i<nprs; i++){
		((PNNCFileRecord *)prs[i])->pcal_hnd = pcal_hnds[i];
		/* Get pfile meta data using the file record */
		ret = ParNCLNCGetFileMData((PNNCFileRecord *)prs[i]);
		if(ret < NhlNOERROR){
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Reading meta data from file failed");
			return NhlWARNING;
		}
	}

	NclFree(pcal_hnds);

	return NhlNOERROR;
}

static NclQuark *ParNCLNCGetVarNames(void *pr, int *n_vars)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCVarInqRecList *cur_rec = NULL;
	NclQuark *var_names;
	int i=0;

	var_names = (NclQuark *)NclCalloc(prec->n_vars, sizeof(NclQuark));
	if(var_names == NULL){
		NhlPError(NhlFATAL, ENOMEM, NULL);
		return NULL;
	}

	cur_rec = prec->vars;
	for(i = 0; (i < prec->n_vars) && (cur_rec) ; i++){
		/* assert(cur_rec->var_inq != NULL); */
		var_names[i] = cur_rec->var_inq->name;
		cur_rec = cur_rec->next;
	}	
	*n_vars = i;

	return var_names;
}

static NclFVarRec *ParNCLNCGetVarInfo(void *pr, NclQuark var_name)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCVarInqRecList *cur_rec = NULL;
	NclFVarRec *fvar_rec = NULL;
	int i=0;

	cur_rec = prec->vars;
	while(cur_rec != NULL){
		if(cur_rec->var_inq->name == var_name){
			fvar_rec = (NclFVarRec *) NclCalloc(1, sizeof(NclFVarRec));
			
			fvar_rec->var_name_quark = cur_rec->var_inq->name;
			fvar_rec->var_real_name_quark = cur_rec->var_inq->name;
			fvar_rec->var_full_name_quark = cur_rec->var_inq->name;
			fvar_rec->data_type = ParNCLNCMapToNcl((void *)(&(cur_rec->var_inq->data_type)));
			fvar_rec->num_dimensions = cur_rec->var_inq->n_dims;

			for(i = 0; (i < cur_rec->var_inq->n_dims) && (i < NCL_MAX_DIMENSIONS); i++){
				fvar_rec->file_dim_num[i] = cur_rec->var_inq->dim_info[i].dim_id;
			}
		}
		cur_rec = cur_rec->next;
	}
	
	return fvar_rec;
}

static NclQuark *ParNCLNCGetVarAttNames(void *pr, NclQuark var_name, int *n_atts)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCVarInqRecList *cur_rec = NULL;
	NclQuark *att_names = NULL;
	int i=0;

	*n_atts = 0;
	cur_rec = prec->vars;
	while(cur_rec != NULL){
		if(cur_rec->var_inq->name == var_name){
			PNNCAttInqRecList *cur_att = NULL;

			*n_atts = cur_rec->var_inq->n_atts;
			cur_att = cur_rec->var_inq->att_list;
			att_names = (NclQuark *) NclCalloc(*n_atts, sizeof(NclQuark));

			for(i=0; (cur_att) && (i < NCL_MAX_ATTRIBUTES); i++){
				att_names[i] = cur_att->att_inq->name;
				cur_att = cur_att->next;
			}
			break;
		}
		cur_rec = cur_rec->next;
	}
	
	return att_names;
}

static NclFAttRec *ParNCLNCGetVarAttInfo(void *pr, NclQuark var_name, NclQuark att_name)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCVarInqRecList *cur_vrec = NULL;
	PNNCAttInqRecList *cur_arec = NULL;
	NclFAttRec *att_rec = NULL;
	int i=0;

	cur_vrec = prec->vars;
	cur_arec = NULL;
	while(cur_vrec != NULL){
		if(cur_vrec->var_inq->name == var_name){
			cur_arec = cur_vrec->var_inq->att_list;
			break;
		}
		cur_vrec = cur_vrec->next;
	}
	while(cur_arec){
		if(cur_arec->att_inq->name == att_name){
			att_rec = (NclFAttRec *) NclCalloc(1, sizeof(NclFAttRec));
			att_rec->att_name_quark = cur_arec->att_inq->name;
			att_rec->data_type = cur_arec->att_inq->data_type;
			att_rec->num_elements = cur_arec->att_inq->len;
			break;
		}
		cur_arec = cur_arec->next;
	}
	if(!att_rec){
		NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to get variable (%s) attributes info for %s", NrmQuarkToString(var_name), NrmQuarkToString(att_name));
	}
	
	return att_rec;
}

static void *ParNCLNCReadVar(void *pr, NclQuark var_name, long *start, long *finish, long *stride, void *storage)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	void *p_shnd = NULL;
	/* FIXME: Read variables only when needed */
	/* FIXME: We should return a ParCAL datatype handle here */
	/* FIXME: What if the same variable is read twice ? */

	p_shnd = ParCALCacheVar(&(prec->pcal_hnd), 1, var_name, NULL, 0);

	return (p_shnd);
}

static void *ParNCLNCPReadVar(void **prs, int nprs, NclQuark var_name, NclFVarSelRec *sel, int nsel, void *storage)
{
	void **pcal_hnds = NULL;
	void *p_shnd = NULL;
	int i;
	/* FIXME: Read variables only when needed */
	/* FIXME: We should return a ParCAL datatype handle here */
	/* FIXME: What if the same variable is read twice ? */

	if(nsel && (nsel != nprs)){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Reading variable (%s) failed, invalid var selector provided", NrmQuarkToString(var_name));
		return NULL;
	}

	/* FIXME: Avoid these unnecessary mallocs, maybe ParCALCacheVar() should
	 * recognize parallel file records
	 */
	pcal_hnds = NclCalloc(nprs, sizeof(void *));
	if(pcal_hnds == NULL){
		NhlPError(NhlFATAL, ENOMEM, NULL);
		return NULL;
	}

	for(i=0; i<nprs; i++){
		PNNCFileRecord *prec = (PNNCFileRecord *)prs[i];
		((void **)pcal_hnds)[i] = prec->pcal_hnd;
	}

	p_shnd = ParCALCacheVar(pcal_hnds, nprs, var_name, sel, nsel);

	NclFree(pcal_hnds);

	return (p_shnd);
}

static NhlErrorTypes ParNCLNCAddVar(void *pr, NclQuark var_name, NclBasicDataTypes datatype, int n_dims, NclQuark *dim_names, long *dim_sizes)
{
	NhlPError(NhlFATAL, NhlEUNKNOWN, "AddVar() is not supported");
	return NhlFATAL;
}

static NhlErrorTypes ParNCLNCWriteVar(void *pr, NclQuark var_name, void *data, long *start, long *finish, long *stride)
{
	NhlPError(NhlFATAL, NhlEUNKNOWN, "WriteVar() is not supported");
	return NhlFATAL;
}

static NclQuark *ParNCLNCGetDimNames(void *pr, int *n_dims)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCDimInqRecList *cur_rec = NULL;
	NclQuark *dims;
	int i=0;

	dims = (NclQuark *)NclCalloc(prec->n_dims, sizeof(NclQuark));
	if(dims == NULL){
		NhlPError(NhlFATAL, ENOMEM, NULL);
		return NULL;
	}

	cur_rec = prec->dims;
	for(i = 0; (i < prec->n_dims) && (cur_rec) ; i++){
		/* assert(cur_rec->dim_inq != NULL); */
		dims[i] = cur_rec->dim_inq->name;
		cur_rec = cur_rec->next;
	}
	*n_dims = i;

	return dims;
}

static NclFDimRec *ParNCLNCGetDimInfo(void *pr, NclQuark dim_name)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCDimInqRecList *cur_rec = NULL;
	NclFDimRec *fdim_rec = NULL;

	cur_rec = prec->dims;
	while(cur_rec){
		if(cur_rec->dim_inq->name == dim_name){
			fdim_rec = (NclFDimRec *)NclCalloc(1, sizeof(NclFDimRec));
			if(fdim_rec != NULL){
				fdim_rec->dim_name_quark = dim_name;
				fdim_rec->dim_size = cur_rec->dim_inq->size;
				fdim_rec->is_unlimited = cur_rec->dim_inq->is_unlimited;
			}
			else{
				NhlPError(NhlFATAL, ENOMEM, NULL);
			}
			break;
		}
		cur_rec = cur_rec->next;
	}
	
	return fdim_rec;
}

static NclQuark *ParNCLNCGetAttNames(void *pr, int *n_atts)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCAttInqRecList *cur_rec = NULL;
	NclQuark *atts;
	int i=0;

	atts = (NclQuark *)NclCalloc(prec->n_file_atts, sizeof(NclQuark));

	cur_rec = prec->file_atts;
	for(i = 0; (i < prec->n_file_atts) && (cur_rec) ; i++){
		atts[i] = cur_rec->att_inq->name;
		cur_rec = cur_rec->next;
	}
	*n_atts = i;

	return atts;
}

static NclFAttRec *ParNCLNCGetAttInfo(void *pr, NclQuark att_name)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCAttInqRecList *cur_rec = NULL;
	NclFAttRec *fatt_rec = NULL;

	cur_rec = prec->file_atts;
	while(cur_rec){
		if(cur_rec->att_inq->name == att_name){
			fatt_rec = (NclFAttRec *)NclCalloc(1, sizeof(NclFAttRec));
			if(fatt_rec != NULL){
				fatt_rec->att_name_quark = att_name;
				fatt_rec->data_type = cur_rec->att_inq->data_type;
				fatt_rec->num_elements = cur_rec->att_inq->len;
			}
			else{
				NhlPError(NhlFATAL, ENOMEM, NULL);
			}
			break;
		}
		cur_rec = cur_rec->next;
	}
	
	return fatt_rec;
}

static void *ParNCLNCReadAtt(void *pr, NclQuark att_name, void *storage)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCAttInqRecList *cur_rec = NULL;

	cur_rec = prec->file_atts;
	while(cur_rec){
		if(cur_rec->att_inq->name == att_name){
			if(storage != NULL){
				memcpy(storage, cur_rec->att_inq->value, cur_rec->att_inq->len * _NclSizeOf(cur_rec->att_inq->data_type));
			}
			break;
		}
		cur_rec = cur_rec->next;
	}
	
	return storage;
}

static void *ParNCLNCReadVarAtt(void *pr, NclQuark var_name, NclQuark att_name, void *storage)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	PNNCVarInqRecList *cur_vrec = NULL;
	PNNCAttInqRecList *cur_arec = NULL;
	int i=0;

	cur_vrec = prec->vars;
	cur_arec = NULL;
	while(cur_vrec != NULL){
		if(cur_vrec->var_inq->name == var_name){
			cur_arec = cur_vrec->var_inq->att_list;
			break;
		}
		cur_vrec = cur_vrec->next;
	}
	while(cur_arec){
		if(cur_arec->att_inq->name == att_name){
			memcpy(storage, cur_arec->att_inq->value, cur_arec->att_inq->len * _NclSizeOf(cur_arec->att_inq->data_type));
			break;
		}
		cur_arec = cur_arec->next;
	}
	
	return storage;
}

static NclFVarRec *ParNCLNCGetCoordInfo(void *pr, NclQuark var_name)
{
	/* FIXME: ParCAL does not provide this info right now */
	/* NhlPError(NhlFATAL, NhlEUNKNOWN, "GetCoordInfo() is not supported"); */
	return NULL;
}

static void *ParNCLNCReadCoord(void *pr, NclQuark var_name, long *start, long *finish, long *stride, void *storage)
{
	NhlPError(NhlWARNING, NhlEUNKNOWN, "ReadCoord() is not supported");
	return NULL;
}

static NclBasicDataTypes ParNCLNCMapToNcl(void *type)
{
	NclBasicDataTypes ncl_type = NCL_none;
	switch(*((nc_type *) type)){
		case NC_DOUBLE:
			ncl_type = NCL_pdouble;
			break;
		default:
			 NhlPError(NhlFATAL, NhlEUNKNOWN, "MapToNcl() is not supported for type (%d)", *(nc_type *)type);
	}

	return ncl_type;
}

static void *ParNCLNCMapFromNcl(NclBasicDataTypes type)
{
	NhlPError(NhlFATAL, NhlEUNKNOWN, "MapFromNcl() is not supported");
	return NULL;
}

static NhlErrorTypes ParNCLNCSetOption(void *pr, NclQuark opt, NclBasicDataTypes datatype, int n_items, void *values)
{
	PNNCFileRecord *prec = (PNNCFileRecord *)pr;
	NhlErrorTypes ret = NhlNOERROR;

	if(opt == NrmStringToQuark("suppressclose")){
		if((int )(prec->options[PNNC_SUPPRESS_CLOSE_OPT].values) == 0){
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Cannot change suppress close option");
			ret = NhlWARNING;
		}
	}
	else{
		NhlPError(NhlWARNING, NhlEUNKNOWN, "SetOption() is not supported for this option (%s)", NrmQuarkToString(opt));
		ret = NhlWARNING;
	}

	return ret;
}

NclFormatFunctionRec ParNCLNCRec = {
/* NclInitializeFileRecFunc initialize_file_rec */      ParNCLNCInitializeFileRec,
/* NclCreateFileFunc	   create_file; */		ParNCLNCCreateFile,
/* NclPCreateFilesFunc	   pcreate_files; */	ParNCLNCCreateFiles,
/* NclOpenFileFunc         open_file; */		ParNCLNCOpenFile,
/* NclPOpenFilesFunc       popen_files; */		ParNCLNCOpenFiles,
/* NclFreeFileRecFunc      free_file_rec; */		ParNCLNCFreeFileRec,
/* NclGetVarNamesFunc      get_var_names; */		ParNCLNCGetVarNames,
/* NclGetVarInfoFunc       get_var_info; */		ParNCLNCGetVarInfo,
/* NclGetDimNamesFunc      get_dim_names; */		ParNCLNCGetDimNames,
/* NclGetDimInfoFunc       get_dim_info; */		ParNCLNCGetDimInfo,
/* NclGetAttNamesFunc      get_att_names; */		ParNCLNCGetAttNames,
/* NclGetAttInfoFunc       get_att_info; */		ParNCLNCGetAttInfo,
/* NclGetVarAttNamesFunc   get_var_att_names; */	ParNCLNCGetVarAttNames,
/* NclGetVarAttInfoFunc    get_var_att_info; */		ParNCLNCGetVarAttInfo,
/* NclGetCoordInfoFunc     get_coord_info; */		ParNCLNCGetCoordInfo,
/* NclReadCoordFunc        read_coord; */		ParNCLNCReadCoord,
/* NclReadNoStrideCoordFunc	read_coord_ns; */		NULL,
/* NclReadVarFunc          read_var; */			ParNCLNCReadVar,
/* NclPReadVarFunc         pread_var; */		ParNCLNCPReadVar,
/* NclReadNoStrideVarFunc	read_var_ns; */			NULL,
/* NclReadAttFunc          read_att; */			ParNCLNCReadAtt,
/* NclReadVarAttFunc       read_var_att; */		ParNCLNCReadVarAtt,
/* NclWriteCoordFunc       write_coord; */		NULL,
/* NclWriteNoStrideCoordFunc	write_coord_ns; */		NULL,
/* NclWriteVarFunc         write_var; */		ParNCLNCWriteVar,
/* NclWriteNoStrideVarFunc	write_var_ns; */		NULL,
/* NclWriteAttFunc         write_att; */		NULL,
/* NclWriteVarAttFunc      write_var_att; */		NULL,
/* NclAddDimFunc           add_dim; */			NULL,
/* NclAddChunkDimFunc      add_chunk_dim; */		NULL,
/* NclRenameDimFunc        rename_dim; */		NULL,
/* NclAddVarFunc           add_var; */			ParNCLNCAddVar,
/* NclAddVarChunkFunc      add_var_chunk; */		NULL,
/* NclAddVarChunkCacheFunc add_var_chunk_cache; */	NULL,
/* NclSetVarCompressLevelFunc set_var_compress_level; */ NULL,
/* NclAddVarFunc           add_coord_var; */		NULL,
/* NclAddAttFunc           add_att; */			NULL,
/* NclAddVarAttFunc        add_var_att; */		NULL,
/* NclMapFormatTypeToNcl   map_format_type_to_ncl; */	ParNCLNCMapToNcl,
/* NclMapNclTypeToFormat   map_ncl_type_to_format; */	ParNCLNCMapFromNcl,
/* NclDelAttFunc           del_att; */			NULL,
/* NclDelVarAttFunc        del_var_att; */		NULL,
/* NclGetGrpNamesFunc      get_grp_names; */            NULL,
/* NclGetGrpInfoFunc       get_grp_info; */             NULL,
/* NclGetGrpAttNamesFunc   get_grp_att_names; */        NULL, 
/* NclGetGrpAttInfoFunc    get_grp_att_info; */         NULL,
/* NclSetOptionFunc        set_option;  */              ParNCLNCSetOption
};

NclFormatFunctionRecPtr PNNCAddFileFormat 
#if	NhlNeedProto
(void)
#else 
()
#endif
{
	return(&ParNCLNCRec);
}
