/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "ScdMBMesh.hpp"

namespace parncl{

void ScdMBMesh::find_scd_box(void )
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::ScdInterface *scdi;

	moab::EntityHandle mb_mset = mesh_finfo_->get_mb_mset();

	mb_retval = mb_gcore_->query_interface(scdi);
	if((mb_retval != moab::MB_SUCCESS) || !scdi){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to query the MOAB scdi interface");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

	std::vector<moab::ScdBox *> scd_boxes;
	mb_retval = scdi->get_boxes(scd_boxes);
	if(mb_retval != moab::MB_SUCCESS){
		std::string funcname("UNKNOWN");
		mb_gcore_->release_interface(scdi);
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get MOAB Scd boxes");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

    bool found = false;
    for(std::vector<moab::ScdBox *>::iterator box_iter = scd_boxes.begin(); box_iter != scd_boxes.end(); ++box_iter){
        moab::ScdBox *scd_box = *box_iter;
        moab::EntityHandle scd_box_set = scd_box->box_set();
        if(mb_gcore_->contains_entities(mb_mset, &scd_box_set, 1)){
            found = true;
			scd_box_ = scd_box;
			break;
        }
    }

	mb_gcore_->release_interface(scdi);
    if(!found){
		std::string funcname("UNKNOWN");
        NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get box dims for MOAB mesh set");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
    }
}

void ScdMBMesh::init_ijk_info(void )
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

	/* MOAB functionality changes from version to version - no more global_box_dims()
	const int *gijk = scd_box_->global_box_dims();
	if(gijk == NULL){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get MOAB mesh global dims");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}
	gijk_dims_.resize(3);
	gijk_dims_[0] = std::make_pair(gijk[0], gijk[3]);
	gijk_dims_[1] = std::make_pair(gijk[1], gijk[4]);
	gijk_dims_[2] = std::make_pair(gijk[2], gijk[5]);
	*/

	moab::HomCoord h_min = scd_box_->box_min();
	moab::HomCoord h_max = scd_box_->box_max();

	lijk_dims_.resize(3);
	lijk_dims_[0] = std::make_pair(h_min.i(), h_max.i());
	lijk_dims_[1] = std::make_pair(h_min.j(), h_max.j());
	lijk_dims_[2] = std::make_pair(h_min.k(), h_max.k());

	int x_sz, y_sz, z_sz;
	scd_box_->box_size(x_sz, y_sz, z_sz);

	const double *xc, *yc, *zc;
	mb_retval = scd_box_->get_coordinate_arrays(xc, yc, zc);
	if(mb_retval != moab::MB_SUCCESS){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get MOAB coord arrays");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

	lijk_coords_min_max_.resize(3);
	lijk_coords_min_max_[0] = std::make_pair(xc[0], xc[x_sz - 1]);
	lijk_coords_min_max_[1] = std::make_pair(yc[0], yc[y_sz - 1]);
	lijk_coords_min_max_[2] = std::make_pair(zc[0], zc[z_sz - 1]);
}

ScdMBMesh::ScdMBMesh(moab::Interface *mb_gcore, MPI_Comm mb_comm, moab::ParallelComm *mb_pcomm, MeshFInfo *mesh_finfo, const std::string &part_method):MBMesh::MBMesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method)
{
	find_scd_box();
	init_ijk_info();
}

ScdMBMesh::~ScdMBMesh()
{
}

bool ScdMBMesh::ctest(const std::vector<size_t>& var_dims)
{
	bool ctest_result = true;
	std::vector<size_t>::const_reverse_iterator var_dims_iter = var_dims.rbegin();
	std::vector<std::pair<size_t, size_t> >::const_iterator gijk_dims_iter = gijk_dims_.begin();
	while((var_dims_iter != var_dims.rend()) &&
			(gijk_dims_iter != gijk_dims_.end())){
		std::pair<size_t, size_t> dims = *gijk_dims_iter;
		if((dims.second - dims.first) != *var_dims_iter){
			ctest_result = false;
			break;
		}
		++var_dims_iter;
		++gijk_dims_iter;
	}
	if((gijk_dims_iter != gijk_dims_.end()) && (ctest_result == true)){
		for(; gijk_dims_iter != gijk_dims_.end(); ++gijk_dims_iter){
			std::pair<size_t, size_t> dims = *gijk_dims_iter;
			if(dims.second != 0){
				ctest_result = false;
				break;
			}
		}	
	}

	return ctest_result;
}

bool ScdMBMesh::ctest(ParCALMeshVal &old_val, NclSelectionRecord *sel)
{
	/*
	std::string funcname("UNKNOWN");
	NhlPError(NhlFATAL, NhlEUNKNOWN, "Not implemented");
	throw pargal::pargal_except(std::string("Not implemented"), __FILE__, __LINE__, funcname);
	*/
	return false;
}

void ScdMBMesh::dbg_print_scd_box(moab::ScdBox *scd_box) const
{
}
} // namespace parncl
